/*
    Copyright (c) 2006 Peter Persson
    This code uses a C2P-routine originally written by Mikael Kalms
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    Changelog
    2006-02-04	Created		(Peter Persson)		
*/


	.text
	.globl	_OpenVIDEL16
	.globl	_CloseVIDEL16
	.globl	_UpdateVIDEL16
	.globl	_PaletteVIDEL16
	
/*****************************************************************
*
* long OpenDRIVER();
*
*/

_OpenVIDEL16:
	moveml	d1-d7/a0-a6,sp@-

	movel	#init,-(sp)
	movew	#0x26,-(sp)
	trap	#14
	addl	#6,sp

	moveml	sp@+,d1-d7/a0-a6
	
	movel	screen,d0
	rts


/* Initialisation stuff, supervisor mode */

init:	/* store 256-colour palette */
	movel	#old_pal,-(sp)
	movew	#256,-(sp)
	movew	#0,-(sp)
	movew	#0x5e,-(sp)
	trap	#14
	addl	#10,sp

	/* store STE-compatible palette registers */
	moveml	0xFFFF8240,d0-d7
	moveml	d0-d7,old_pal16
	
	/* store screen address */
	movew	#0x02,-(sp)
	trap	#14
	addql	#2,sp
	movel	d0,old_physbase

	/* store old VIDEL settings */
	movel	#old_mode,a0
	movel	0xFFFF8282,(a0)+
	movel	0xFFFF8286,(a0)+
	movel	0xFFFF828A,(a0)+
	movel	0xFFFF82A2,(a0)+
	movel	0xFFFF82A6,(a0)+
	movel	0xFFFF82AA,(a0)+
	movew	0xFFFF820A,(a0)+
	movew	0xFFFF82C0,(a0)+
	movew	0xFFFF8266,(a0)+
	movew	0xFFFF82C2,(a0)+
	movew	0xFFFF8210,(a0)+

	/* get monitor type */
	movew	#0x59,-(sp)
	trap	#14
	addql	#2,sp
	
	btst	#0,d0	/* primitive, doesn't handle mono monitors etc... */
	bne	rgb	

vga:	movel	#vga_mode,a0	/* force screenmode for vga */
	bra	cont
rgb:	movel	#rgb_mode,a0	/* force screenmode for rgb */

	/* force new VIDEL settings */
cont:	movel	(a0)+,0xFFFF8282
	movel	(a0)+,0xFFFF8286
	movel	(a0)+,0xFFFF828A
	movel	(a0)+,0xFFFF82A2
	movel	(a0)+,0xFFFF82A6
	movel	(a0)+,0xFFFF82AA
	movew	(a0)+,0xFFFF820A
	movew	(a0)+,0xFFFF82C0
	clrw	0xFFFF8266
	movew	(a0)+,0xFFFF8266
	movew	(a0)+,0xFFFF82C2
	movew	(a0)+,0xFFFF8210
	
	/* reserve video memory in ST-RAM */	
	movew	#0,-(sp)		| ST-ram only
	movel	#100000,-(sp)	| Size of 1 * 256*192*16bpp + some extra bytes for alignment
	movew	#0x44,-(sp)
	trap	#1
	addql	#8,sp
	
	addl	#256,d0
	andl	#0xffffff00,d0	/* align the data */
	movel	d0,screen

	/* set screen address */
	
	movel	d0,0xff8200	/* msb */
	movew	d0,0xff8203	/* middle byte */
	moveb	d0,0xff820D	/* lsb */ 

	/* set border colour to black */
	movel	#0,0xffff9800

	/* clear frameskip counter */
	movel	#0,flipcntr

	/* install VBI */
	movew	#0x2700,sr
	movel	0x70,old_vbi
	movel	#vbi,0x70
	movew	#0x2300,sr
	rts


/*****************************************************************
*
* Interrupt routine (internal);
*
*/

	.text
	.even
	.ascii	"XBRA"
	.ascii	"SMSP"
	.comm	old_vbi,4

vbi:	/* handle frameskip counter */
	tstl	flipcntr
	beq	vbi_done

	subl	#1,flipcntr


vbi_done:	movel	old_vbi,-(sp)	/* jump through old vector */
	rts

	
	.comm	vbi_saveregs,64


/*****************************************************************
*
* void CloseDRIVER();
*
*/


_CloseVIDEL16:
	moveml	d0-d7/a0-a6,sp@-

	movel	#exit,-(sp)
	movew	#0x26,-(sp)
	trap	#14
	addl	#6,sp

	moveml	sp@+,d0-d7/a0-a6
	moveq	#0,d0		/* alles OK */
	rts

exit:	/* restore VBI */
	movew	#0x2700,sr
	movel	old_vbi,0x70
	movew	#0x2300,sr
	
	/* restore old videomode */
	movel	#old_mode,a0
	movel	(a0)+,0xFFFF8282
	movel	(a0)+,0xFFFF8286
	movel	(a0)+,0xFFFF828A
	movel	(a0)+,0xFFFF82A2
	movel	(a0)+,0xFFFF82A6
	movel	(a0)+,0xFFFF82AA
	movew	(a0)+,0xFFFF820A
	movew	(a0)+,0xFFFF82C0
	clrw	0xFFFF8266
	movew	(a0)+,0xFFFF8266
	movew	(a0)+,0xFFFF82C2
	movew	(a0)+,0xFFFF8210

	/* restore STE-compatible palette registers */
	moveml	old_pal16,d0-d7
	moveml	d0-d7,0xFFFF8240

	/* restore 256-colour palette */
	movel	#old_pal,-(sp)
	movew	#256,-(sp)
	movew	#0,-(sp)
	movew	#0x5d,-(sp)
	trap	#14
	addl	#10,sp

	/* restore old screen address */

	movew	#-1,-(sp)
	movel	old_physbase,-(sp)
	movel	old_physbase,-(sp)
	movew	#0x05,-(sp)
	trap	#14
	addl	#12,sp

	/* release memory */

	movel	screen,-(sp)
	movew	#0x49,-(sp)
	trap	#1
	addql	#6,sp
	
	rts

/*****************************************************************
*
* UpdateDRIVER(source_bitmap);
*
*/

_UpdateVIDEL16:
	tstl	flipcntr
	bne	_UpdateVIDEL16

	movel	8(sp),flipcntr
	rts
	
/*****************************************************************
*
* PaletteDRIVER(source_pal);
*
*/

_PaletteVIDEL16:
	/* do nothing, no palette handling needed */
	rts
	
	.data
	.even
	.comm	flipcntr,4
	.comm	screen,4
	.comm	old_physbase,4


/*****************************************************************
*
* Videomodes;
*
*/

	/* 256x192x16bpp, 60hz, VGA */
	/* uses 25Mhz clock */
vga_mode:	dc.l	0xC6008D, 0x1502B9, 0x5A0097, 0x41903FF, 0x3F00A1, 0x3A10415
	dc.w	0x200, 0x186, 0x100, 0x5, 0x100

	/* 256x192x16bpp, 60hz, RGB */
	/* uses 25Mhz clock */
rgb_mode:	dc.l	0xC7009F, 0x1F0022, 0x7A00AB, 0x20D0201, 0x17004B, 0x1CB0207
	dc.w	0x200, 0x187, 0x100, 0x0, 0x100


	/* holds saved VIDEL register settings */
old_mode:	ds.l	9
old_pal16:ds.l	8
old_pal:	ds.l	1024
